#include <SDL.h>

#include "keypad.h"

#define DG_KEY_A SDLK_LCTRL
#define DG_KEY_B SDLK_LALT
#define DG_KEY_X SDLK_SPACE
#define DG_KEY_Y SDLK_LSHIFT

#define DG_KEY_L SDLK_TAB
#define DG_KEY_R SDLK_BACKSPACE
#define DG_KEY_START  SDLK_RETURN
#define DG_KEY_SELECT SDLK_ESCAPE

#define DG_KEY_UP    SDLK_UP
#define DG_KEY_DOWN  SDLK_DOWN
#define DG_KEY_LEFT  SDLK_LEFT
#define DG_KEY_RIGHT SDLK_RIGHT

static int pad_quit = 0;

int pad_is_quit(void)
{
	return pad_quit;
}
struct sdl98_key {
	SDLKey sym;
	unsigned char key;
};

struct sdl98_key sdl98_keytbl[] = {
	SDLK_1, 0x01, // ! : pc-98 same as 101key with sfift
	SDLK_2, 0x02,
	SDLK_3, 0x03, // # : pc-98 same as 101key with sfift
	SDLK_4, 0x04,
	SDLK_5, 0x05, // % : pc-98 same as 101key with sfift 
	SDLK_6, 0x06,
	SDLK_7, 0x07,
	SDLK_8, 0x08,
	SDLK_9, 0x09,
	SDLK_0, 0x0a,
	SDLK_a, 0x1d,
	SDLK_b, 0x2d,
	SDLK_c, 0x2b,
	SDLK_d, 0x1f,
	SDLK_e, 0x12,
	SDLK_f, 0x20,
	SDLK_g, 0x21,
	SDLK_h, 0x22,
	SDLK_i, 0x17,
	SDLK_j, 0x23,
	SDLK_k, 0x24,
	SDLK_l, 0x25,
	SDLK_m, 0x2f,
	SDLK_n, 0x2e,
	SDLK_o, 0x18,
	SDLK_p, 0x19,
	SDLK_q, 0x10,
	SDLK_r, 0x13,
	SDLK_s, 0x1e,
	SDLK_t, 0x14,
	SDLK_u, 0x16,
	SDLK_v, 0x2c,
	SDLK_w, 0x11,
	SDLK_x, 0x2a,
	SDLK_y, 0x15,
	SDLK_z, 0x29,
	SDLK_RETURN, 0x1c,
	SDLK_LSHIFT, 0x70,
	SDLK_RSHIFT, 0x7d,
	SDLK_MINUS, 0x0b,
	SDLK_SLASH, 0x32, // ? : pc-98 same as 101key with sfift
	SDLK_SEMICOLON, 0x26, // without shift
	SDLK_PERIOD, 0x31, // > : pc-98 same as 101key with sfift
	SDLK_COMMA, 0x30, // < : pc-98 same as 101key with sfift
	SDLK_LEFTBRACKET, 0x1b, // { : pc-98 same as 101key with sfift
	SDLK_RIGHTBRACKET, 0x28, // } : pc-98 same as 101key with sfift
	SDLK_BACKSLASH, 0x0d, // | : pc-98 same as 101key with sfift
	SDLK_QUOTE, 0x07,
	SDLK_EQUALS, 0x0b,
	0xff, 0xff, // end of tbl
};

/*shiftt: :()$&@"?!{}#%^*+_|~<> */
struct sdl98_key sdl98_shift_keytbl[] = {
//	SDLK_SEMICOLON, 0x26, // :
	SDLK_9, 0x08, // (
	SDLK_0, 0x09, // )
	SDLK_2, 0x1a, // @
	SDLK_4, 0x04, // $
	SDLK_7, 0x06, // &
	SDLK_6, 0x0c, // ^
	SDLK_8, 0x27, // *
	SDLK_QUOTE, 0x02, // "
	SDLK_EQUALS, 0x26, // +
	SDLK_BACKQUOTE, 0x0c, // ~
	SDLK_MINUS, 0x33, // _
	SDLK_SEMICOLON, 0x27, // :
	0xff, 0xff, // end of tbl
};

static void sendkey(unsigned char key, int shift)
{
	if (shift) {
		keystat_senddata(0x70); // LSHIFT
	}
	keystat_senddata(key);
	keystat_senddata(key | 0x80);
	if (shift) {
		keystat_senddata(0x70 | 0x80); // LSHIFT
	}
}

/* ̂܂܁AVtgtł̂܂܁AVtgtŕϊ */
void conv_sdl_98key(SDLKey k, int shift)
{
	int i;
	struct sdl98_key *tbl;

	if (shift) {
		tbl = sdl98_shift_keytbl;
		for (i = 0; tbl[i].key != 0xff; i++) {
			if (tbl[i].sym == k) {
				sendkey(tbl[i].key, (k == SDLK_SEMICOLON || k == SDLK_2)?0:1);
				return;
			}
		}
	}

	tbl = sdl98_keytbl;
	for (i = 0; tbl[i].key != 0xff; i++) {
		if (tbl[i].sym == k) {
			sendkey(tbl[i].key, (k == SDLK_QUOTE || k == SDLK_EQUALS)?1:shift);
			return;
		}
	}
}

unsigned int pad_event(SDL_Event *evt, int *x, int *y, int *dx, int *dy)
{
	static int mouse_up_flag = 1, shiftkey_down_flag = 0;
	int pad_data = 0;

	if (evt->type == SDL_QUIT)
	{
		pad_quit = 1;
	}
	else if (evt->type == SDL_KEYDOWN) {
		printf("keydown:%d", evt->key.keysym.sym);
		if (evt->key.keysym.sym == SDLK_LSHIFT) {
			shiftkey_down_flag = 1;
		} else {
			conv_sdl_98key(evt->key.keysym.sym, shiftkey_down_flag);
			shiftkey_down_flag = 0;
		}
		if (evt->key.keysym.sym == SDLK_RETURN) {
			pad_data |= PAD_R;
		}
	}
	else if (evt->type == SDL_MOUSEBUTTONDOWN) {
		SDL_GetMouseState(x, y);
	}
	else if (evt->type == SDL_MOUSEMOTION) {
	  	SDL_GetMouseState(x, y);
		SDL_GetRelativeMouseState(dx, dy);
		if (mouse_up_flag) {
			mouse_up_flag = 0;
			*dx = *dy = 0; // clear
		}
	} else if (evt->type == SDL_MOUSEBUTTONUP) {
		mouse_up_flag = 1;
	}
	
	return pad_data;
}

unsigned int pad_poll(int *x, int *y, int *dx, int *dy)
{
	SDL_Event evt;
	int pad_data = 0;

	while(SDL_PollEvent(&evt) && !pad_quit)
	{
		pad_data = pad_event(&evt, x, y, dx, dy);
	}
	
	return pad_data;
}


#ifdef PAD_TEST

SDL_Surface *gMain;

int pad_init(void)
{
	if (SDL_Init(SDL_INIT_VIDEO))
		return -1;

	gMain = SDL_SetVideoMode(320,240,16,0);
	
	if (!gMain)
		return -1;
}

void pad_release(void)
{
	SDL_Quit();
}

int main(int argc,char *argv[])
{
	unsigned data,olddata;
	
	pad_init();
	
	olddata = 0;
	
	while(!pad_quit)
	{
		do
		{
			SDL_Delay(10);
			data = pad_poll();
		}while(!pad_quit && olddata == data);
		
		printf("data = %x\n",data);
		
		olddata = data;
	}
	
	pad_release();
	
	return 0;
}

#endif
